/* TCP/UDP relay daemon */
/* SHUDOH Kazuyuki  Nov 7 1996 */

#include <stdio.h>
#include <stdlib.h>	/* for atol() */
#include <string.h>	/* for str*(), mem*() */
#include <unistd.h>	/* for getopt() */

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/* for select(2) */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "commtcp.h"
#include "commudp.h"

#define TCP_COMMAND	"fwtcp"
#define TCP_COMMAND_LEN	5
#define UDP_COMMAND	"fwudp"
#define UDP_COMMAND_LEN	5

#define MAX_OUT		16
#define MAX_ARG_LEN	256
#define BUF_LEN		1024


void fatal();
void usage(char *command);

int flag_verbose = 0;	/* false */

int in_port;
char *out_host[MAX_OUT];
int out_port[MAX_OUT];
struct sockaddr_in out_sockaddr[MAX_OUT];
int out_sock[MAX_OUT];

int lsock, sock;
fd_set zero_fds;
fd_set read_fds;


int main(int argc, char **argv) {
  int c;	/* for getopt() */
  extern char *optarg;	/* for getopt() with SunOS4 */
  struct sockaddr me, from;
  int read_len;
  int i;
  char buf[BUF_LEN];

  FD_ZERO(&zero_fds);


  while ((c = getopt(argc, argv, "hv")) >= 0) {
    switch (c) {
    case 'h':
      usage(argv[0]);
      exit(0);
    case 'v':
      flag_verbose = 1;	/* true */
      break;
    default:
      break;
    }
  }


  if (argc < 3) {
    usage(argv[0]);  exit(1);
  }

  in_port = (int)atol(argv[1]);
  if ((in_port <= 0) || (in_port >= 0x10000)) {
    printf("Invalid port no: %d\n", in_port);
    usage(argv[0]);  exit(1);
  }

  /* ۥ̾:ݡֹΥѡ */
  for (i = 0; i < argc - 2; i++) {
    char hostname[MAX_ARG_LEN];
    char *ptr;

    /* ۥ̾ */
    strcpy(hostname, argv[i+2]);
    ptr = hostname;
    while (*ptr != '\0') {
      if (*ptr == ':') {
	*ptr = '\0';
	break;
      }
      ptr++;
    }
    out_host[i] = (char *)strdup(hostname);

    /* ݡֹ */
    out_port[i] = (int)atol(++ptr);

    if ((out_port[i] <= 0) || (out_port[i] >= 0x10000)) {
      printf("Invalid port no: %d\n", out_port[i]);
      usage(argv[0]);  exit(1);
    }

#ifdef DEBUG
    printf("hostname: %s, portno: %d\n", out_host[i], out_port[i]);
    fflush(stdout);
#endif /* DEBUG */

    if (SetSockAddr(&(out_sockaddr[i]), out_host[i], out_port[i]) < 0) {
      exit(1);
    }
  }

  /* terminate arrays */
  out_host[i] = NULL;
  out_port[i] = -1;
  out_sock[i] = -1;


  /* listen */
  if (SetSockAddr((struct sockaddr_in *)&me, NULL, in_port) < 0) {
    printf("SetSockAddr failed.\n");
    exit(1);
  }


  /* ޥɤξʬ */
  /* UDP */
  if (!strncmp(argv[0] + strlen(argv[0]) - UDP_COMMAND_LEN,
	       UDP_COMMAND, UDP_COMMAND_LEN)) {
    if ((sock = UdpSetup((struct sockaddr_in *)&me)) < 0) {
      printf("UdpSetup failed.\n");
      exit(1);
    }

    while (1) {
      int fromlen, acptlen, tolen;

      /*  */
      fromlen = sizeof(from);	/* ɬ */
      acptlen = recvfrom(sock, buf, BUF_LEN,
			0, (struct sockaddr *)&from, &fromlen);
      if (acptlen < 0) {
	perror("recvfrom");
      }

      if (flag_verbose) {
	printf("received `%s'\n", buf);
      }

      /*  */
      i = 0;
      while (out_host[i] != NULL) {
	tolen = sizeof(out_sockaddr[i]);

	if (sendto(sock, buf, acptlen,
		0, (struct sockaddr *)&out_sockaddr[i], tolen) != acptlen)
	  fprintf(stderr, "send error\n");

	i++;
      }
    }
  }
  /* TCP */
  else if (!strncmp(argv[0] + strlen(argv[0]) - TCP_COMMAND_LEN,
		    TCP_COMMAND, TCP_COMMAND_LEN)) {
    if ((lsock = TcpSetup((struct sockaddr_in *)&me, 5)) < 0) {
      printf("TcpSetup failed.\n");
      exit(1);
    }

    /* main loop */
restart:
    while (1) {
      int nfds;
      int nbyte;
      fd_set error_fds;
      fd_set write_fds;

      if ((sock = accept(lsock, &from, &nbyte)) < 0) {
	perror("accept");
	continue;
      }

      /* connect */
      i = 0;
      while (out_host[i]) {
#ifdef DEBUG
	printf("connecting to %s:%d\n", out_host[i], out_port[i]);
#endif
	out_sock[i] = TcpConnect(&(out_sockaddr[i]));


	if ((out_sock[i] < 0) && (i == 0)) {
	  fatal();  goto restart;
	}

	i++;
      }


      /* ɤ߽񤭥롼 */
      while (1) {
	int width = -1;

	FD_ZERO(&read_fds);
	FD_ZERO(&write_fds);
	FD_ZERO(&error_fds);

	FD_SET(sock, &read_fds);
	if (sock >= width)  width = sock + 1;

	FD_SET(out_sock[0], &read_fds);
	if (out_sock[0] >= width)  width = out_sock[0] + 1;


	nfds = select(width, &read_fds, &write_fds, &error_fds, NULL);
	if (nfds < 0) {
	  fatal();  goto restart;
	}

	/* read and write */
	if (FD_ISSET(sock, &read_fds)) {
	  read_len = BUF_LEN;

	  while (read_len == BUF_LEN) {
	    if ((read_len = read(sock, buf, BUF_LEN)) < 0) {
	      fatal();  goto restart;
	    }
	    else if (read_len == 0) {	/* closed */
	      fatal();  goto restart;
	    }

	    i = 0;
	    while (out_sock[i] >= 0) {
	      int rest = read_len;
	      int wrote_len;
	      char *ptr = buf;

	      while (rest > 0) {
		wrote_len = write(out_sock[i], ptr, rest);
		rest -= wrote_len;
		ptr += wrote_len;
	      }
	      i++;
	    }
	  }
	}

	if (FD_ISSET(out_sock[0], &read_fds)) {
	  read_len = BUF_LEN;

	  while (read_len == BUF_LEN) {
	    int rest;
	    int wrote_len;
	    char *ptr;

	    if ((read_len = read(out_sock[0], buf, BUF_LEN)) < 0) {
	      fatal();  goto restart;
	    }
	    else if (read_len == 0) {	/* closed */
	      fatal();  goto restart;
	    }

	    rest = read_len;
	    ptr = buf;
	    while (rest > 0) {
	      wrote_len = write(sock, ptr, rest);
	      rest -= wrote_len;
	      ptr += wrote_len;
	    }
	  }
	}

	if (memcmp(&error_fds, &zero_fds, sizeof(fd_set))) {
#ifdef DEBUG
	  printf("restart.\n");
#endif
	  fatal();  goto restart;
	}
      }	/* ɤ߽񤭥롼 */
    }
  }	/* ޥɤξʬ */

  return 0;
}

void fatal() {
  int i;

  close(sock);

  i = 0;
  while (out_sock[i] >= 0) {
    close(out_sock[i]);
    i++;
  }
}

void usage(char *cmd) {
  printf("usage: %s port hostname:port [hostname:port]...\n", cmd);
}
